DEFAULT_TASK_TIMEOUT = 60 * 60 * 24  # 1 day
EXPIRING_REGISTRIES_TTL = 60 * 60 * 24 * 7  # 7 days
DEAD_WORKER_TTL = 60 * 60  # 1 hour
WORKER_HEARTBEAT_FREQ = 10  # 10 seconds
WORKER_HEARTBEAT_TIMEOUT = 60  # 1 minute
MAINTENANCE_FREQ = 60 * 1  # 1 minute

REDIS_URL = "redis://127.0.0.1:6379"
REDIS_PREFIX = "redis_tasks"
MIDDLEWARE = []
WORKER_PRELOAD_FUNCTION = None
WORKER_DESCRIPTION_FUNCTION = "redis_tasks.worker_process.generate_worker_description"

# SCHEDULE is an id -> entry dict. The entries are dicts with the following keys:
#   task: Import path of the function to be run, e.g. "mymodule.myfunc"
#   schedule: Schedule for this task, e.g. crontab("2 4 * * mon,fri")
#   args, kwargs: args and kwargs for the tak function (optional)
#   singleton: Boolean to specify whether this task should not be on the queue
#              multiple times (optional, defaults to True)
#   queue: Queue to put this task on (optional
SCHEDULE = {}
SCHEDULER_TIMEZONE = "UTC"
SCHEDULER_QUEUE = 'default'
SCHEDULER_MAX_CATCHUP = 60 * 60 * 1  # 1 hour



import os.path

import click

from redis_tasks.cli import main as cli_main


class Command:
    def run_from_argv(self, argv):
        """Called when run from the command line."""
        prog_name = "{} {}".format(os.path.basename(argv[0]), argv[1])
        return self.main(args=argv[2:], prog_name=prog_name)

    def print_help(self, prog_name, subcommand):
        prog_name = "{} {}".format(prog_name, subcommand)
        return self.main(['--help'], prog_name=prog_name)

    def main(self, args, prog_name):
        decorators = [
            click.option(
                '--pythonpath',
                metavar='PYTHONPATH',
                expose_value=False,
                help=('A directory to add to the Python path, e.g. '
                      '"/home/djangoprojects/myproject".'),
            ),
            click.option(
                '--settings',
                metavar='SETTINGS',
                expose_value=False,
                help=('The Python path to a settings module, e.g. '
                      '"myproject.settings.main". If this is not provided, the '
                      'DJANGO_SETTINGS_MODULE environment variable will be used.'),
            ),
        ]
        command = cli_main
        for decorator in decorators:
            command = decorator(command)
        return command.main(args=args, prog_name=prog_name)



import os

from ..worker import generate_worker_description as default_description


def generate_worker_description(*args, **kwargs):
    if 'DYNO' in os.environ:
        return os.environ['DYNO']
    else:
        return default_description(*args, **kwargs)



from itertools import chain

from django.apps import AppConfig
from django.conf import settings as django_settings

from ...conf import settings
from ... import defaults

SETTINGS_PREFIX = 'RT_'


class RTDjango(AppConfig):
    name = 'redis_tasks.contrib.django'
    label = 'redis_tasks'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        middleware = list(getattr(django_settings, SETTINGS_PREFIX + 'MIDDLEWARE',
                                  defaults.MIDDLEWARE))
        if any(x.startswith('raven.') for x in django_settings.INSTALLED_APPS):
            mw_path = 'redis_tasks.contrib.sentry.SentryMiddleware'
            if mw_path not in middleware:
                middleware.insert(0, mw_path)

        settings.configure(DjangoSettingsProxy(dict(
            SENTRY_INSTANCE="raven.contrib.django.models.client",
            SCHEDULER_TIMEZONE=django_settings.TIME_ZONE,
            MIDDLEWARE=middleware,
        )))


class DjangoSettingsProxy:
    def __init__(self, fallbacks):
        self.fallbacks = fallbacks

    def __getattr__(self, name):
        if hasattr(django_settings, SETTINGS_PREFIX + name):
            return getattr(django_settings, SETTINGS_PREFIX + name)
        elif name in self.fallbacks:
            return self.fallbacks[name]
        else:
            raise AttributeError(name)

    def __dir__(self):
        return set(chain((x[len(SETTINGS_PREFIX):]
                          for x in dir(django_settings)
                          if x.startswith(SETTINGS_PREFIX)),
                         self.fallbacks.keys()))



import redis_tasks

from .conf import RedisKey, connection, settings
from .exceptions import WorkerDoesNotExist
from .utils import atomic_pipeline, decode_list


class ExpiringRegistry:
    def __init__(self, name):
        self.key = RedisKey(name + '_tasks')

    @atomic_pipeline
    def add(self, task, *, pipeline):
        timestamp = connection.ftime()
        pipeline.zadd(self.key, {task.id: timestamp})

    def get_task_ids(self, offset=0, length=-1):
        end = offset + length if length >= 0 else length
        return decode_list(connection.zrange(self.key, offset, end))

    def get_tasks(self, offset=0, length=-1):
        return [redis_tasks.task.Task.fetch(x) for x in self.get_task_ids(offset, length)]

    def empty(self):  # TODO: test
        def transaction(pipeline):
            task_ids = decode_list(pipeline.zrange(self.key, 0, -1))
            pipeline.multi()
            pipeline.delete(self.key)
            redis_tasks.task.Task.delete_many(task_ids, pipeline=pipeline)
        connection.transaction(transaction, self.key)

    def count(self):
        return connection.zcard(self.key)

    def expire(self):
        """Remove expired tasks from registry."""
        timestamp = connection.ftime()
        cutoff_time = timestamp - settings.EXPIRING_REGISTRIES_TTL

        def transaction(pipeline):
            expired_task_ids = decode_list(pipeline.zrangebyscore(
                self.key, 0, cutoff_time))
            if expired_task_ids:
                pipeline.multi()
                pipeline.zremrangebyscore(self.key, 0, cutoff_time)
                redis_tasks.task.Task.delete_many(expired_task_ids, pipeline=pipeline)
        connection.transaction(transaction, self.key)


finished_task_registry = ExpiringRegistry('finished')
failed_task_registry = ExpiringRegistry('failed')


def registry_maintenance():
    finished_task_registry.expire()
    failed_task_registry.expire()
    worker_registry.handle_died_workers()


class WorkerRegistry:
    def __init__(self):
        self.key = RedisKey('workers')

    @atomic_pipeline
    def add(self, worker, *, pipeline):
        timestamp = connection.ftime()
        pipeline.zadd(self.key, {worker.id: timestamp})

    def heartbeat(self, worker):
        timestamp = connection.ftime()
        updated = connection.zadd(self.key, {worker.id: timestamp}, xx=True, ch=True)
        if not updated:
            raise WorkerDoesNotExist()

    @atomic_pipeline
    def remove(self, worker, *, pipeline):
        pipeline.zrem(self.key, worker.id)

    def get_worker_ids(self):
        return decode_list(connection.zrangebyscore(
            self.key, '-inf', '+inf'))

    def get_running_tasks(self):
        """Returns a worker_id -> task_id dict"""
        task_key_prefix = RedisKey('worker_task:')
        lua = connection.register_script("""
            local workers_key, task_key_prefix = unpack(KEYS)
            local worker_ids = redis.call("ZRANGE", workers_key, 0, -1)
            local out = {}
            for _, worker_id in ipairs(worker_ids) do
                local task_key = task_key_prefix .. worker_id
                local task_id = redis.call("LINDEX", task_key, 0)
                if task_id ~= false then
                    table.insert(out, worker_id)
                    table.insert(out, task_id)
                end
            end
            return out
        """)
        it = iter(decode_list(lua(keys=[self.key, task_key_prefix])))
        return dict(zip(it, it))

    def handle_died_workers(self):  # TODO: Test
        from redis_tasks.worker import Worker
        died_worker_ids = self.get_dead_ids()
        for worker_id in died_worker_ids:
            worker = Worker.fetch(worker_id)
            worker.died()

    def get_dead_ids(self):
        oldest_valid = connection.ftime() - settings.WORKER_HEARTBEAT_TIMEOUT
        return decode_list(connection.zrangebyscore(
            self.key, '-inf', oldest_valid))


worker_registry = WorkerRegistry()


class QueueRegistry:
    def __init__(self):
        self.key = RedisKey('queues')

    def get_names(self):
        return decode_list(connection.smembers(self.key))

    @atomic_pipeline
    def add(self, queue, *, pipeline):
        pipeline.sadd(self.key, queue.name)

    @atomic_pipeline
    def remove(self, queue, *, pipeline):
        pipeline.srem(self.key, queue.name)


queue_registry = QueueRegistry()



import logging
import multiprocessing
import os
import signal
import socket
import sys
import threading
import traceback
import uuid
from contextlib import ExitStack as nullcontext
from contextlib import contextmanager
from datetime import timedelta

from .conf import RedisKey, connection, settings, task_middleware
from .exceptions import WorkerShutdown
from .queue import Queue
from .registries import registry_maintenance
from .utils import import_attribute, utcformat, utcnow, utcparse
from .worker import Worker

logger = logging.getLogger('redis_tasks.worker')


class PostponeShutdown:
    _active = set()
    _shutdown_delayed = False

    def activate(self):
        self.__enter__()

    def deactivate(self):
        self.__exit__()

    def __enter__(self):
        self.assert_main_thread()
        self._active.add(self)

    def __exit__(self, *args):
        self.assert_main_thread()
        self._active.remove(self)

        if not self._active and self._shutdown_delayed:
            logger.warning('PostponeShutdown left, raising WorkerShutdown')
            raise WorkerShutdown()

    @staticmethod
    def assert_main_thread():
        if threading.current_thread() != threading.main_thread():
            raise RuntimeError("PostponeShutdown can only be used in main thread")

    @classmethod
    def trigger_shutdown(cls):
        cls.assert_main_thread()
        if cls._active:
            logger.warning('Delaying WorkerShutdown till PostponeShutdown is finished')
            cls._shutdown_delayed = True
        else:
            logger.warning('Raising WorkerShutdown to cancel task')
            raise WorkerShutdown()


class ShutdownRequested(BaseException):
    pass


def generate_worker_description():
    hostname = socket.gethostname()
    shortname = hostname.split('.', maxsplit=1)[0]
    return '{0}.{1}'.format(shortname, os.getpid())


def worker_main(queue_names=["default"], *, burst=False, description=None):
    if isinstance(queue_names, str):
        queue_names = [queue_names]
    process = WorkerProcess([Queue(n) for n in queue_names], description=description)
    try:
        return process.run(burst)
    except ShutdownRequested:
        sys.exit()


class WorkerProcess:
    def __init__(self, queues, *, description=None):
        if not description:
            description_generator = import_attribute(settings.WORKER_DESCRIPTION_FUNCTION)
            description = description_generator()
        id = str(uuid.uuid4())
        self.worker = Worker(id, description=description, queues=queues)
        self.maintenance = Maintenance()
        self.in_interruptible = 0
        self.shutdown_requested = False

    def run(self, burst=False):
        """Starts the work loop.

        Returns the number of tasks that were processed in burst mode"""
        self.install_signal_handlers()
        self.worker.startup()

        if settings.WORKER_PRELOAD_FUNCTION:
            worker_preload = import_attribute(settings.WORKER_PRELOAD_FUNCTION)
            worker_preload(self.worker)

        # Basic test of the middleware setup. If this fails, we would rather
        # have the worker fail to start than to fail every job.
        [x() for x in task_middleware]

        try:
            tasks_processed = 0
            for task in self.queue_iter(burst):
                self.process_task(task)
                tasks_processed += 1
                self.maybe_shutdown()

                self.worker.heartbeat()
                self.maintenance.run_if_neccessary()
                self.maybe_shutdown()
            else:
                logger.info(f"Burst finished after {tasks_processed} tasks, shutting down")
                return tasks_processed
        finally:
            self.worker.shutdown()

    def queue_iter(self, burst):
        logger.debug('Worker listening on {}'.format(
            ', '.join(q.name for q in self.worker.queues)))

        while True:
            self.maybe_shutdown()
            self.worker.heartbeat()
            task = None
            # The queue unblocker might loose entries on worker shutdown, so we
            # regularly try to dequeue unconditionally
            for queue in self.worker.queues:
                self.maybe_shutdown()
                task = queue.dequeue(self.worker)
                if task:
                    break
            if not task and not burst:
                with self.interruptible():
                    queue = Queue.await_multi(self.worker.queues, settings.WORKER_HEARTBEAT_FREQ)
                    if not queue:
                        continue
                # First, attempt to dequeue from the queue whose unblocker we
                # consumed. This makes for more sensible behavior in situations
                # with multiple queues and workers.
                task = queue.dequeue(self.worker)
            if task:
                yield task
            elif burst:
                break

    def process_task(self, task):
        self.worker.start_task(task)
        try:
            outcome = self.execute_task(task)
        except Exception:
            # TODO: check whether this is necessary, execute_tasks should not raise
            exc_string = ''.join(traceback.format_exception(*sys.exc_info()))
            outcome = task.get_abort_outcome(exc_string)
        self.worker.end_task(task, outcome)
        return True

    def execute_task(self, task):
        timeout_at = utcnow() + timedelta(seconds=task.timeout)
        horse_connection, writer = multiprocessing.Pipe(duplex=False)
        work_horse = WorkHorse(task, writer)
        work_horse.daemon = True
        try:
            self.worker.heartbeat()
            work_horse.start()
            # Wait for horse to set up its signal handling
            if horse_connection.poll(5):
                assert horse_connection.recv()
            else:
                logger.error('Workhorse failed to start')
                return task.get_abort_outcome('Workhorse failed to start')
            shutdown_requested = False
            while work_horse.is_alive():
                self.worker.heartbeat()
                if utcnow() >= timeout_at:
                    logger.error('Task reached timeout, killing workhorse')
                    work_horse.send_signal(signal.SIGKILL)
                    return task.get_abort_outcome(f'Task timeout ({task.timeout} sec) reached',
                                                  may_requeue=False)
                try:
                    with self.interruptible() if not shutdown_requested else nullcontext():
                        work_horse.join(settings.WORKER_HEARTBEAT_FREQ)
                except ShutdownRequested:
                    logger.debug('ShutdownRequested caught, signaling WorkHorse to shut down')
                    work_horse.send_signal(signal.SIGUSR1)
                    shutdown_requested = True
        finally:
            if work_horse.is_alive():
                work_horse.send_signal(signal.SIGKILL)
                work_horse.join()

        if horse_connection.poll():
            return horse_connection.recv()
        else:
            logger.error('Workhorse died unexpectedly')
            return task.get_abort_outcome('Workhorse died unexpectedly')

    def install_signal_handlers(self):
        signal.signal(signal.SIGINT, self.handle_stop_signal)
        signal.signal(signal.SIGTERM, self.handle_stop_signal)

    def handle_stop_signal(self, signum, frame):
        logger.debug('Got signal {0}'.format(signum))
        if self.shutdown_requested:
            return
        logger.debug('Shutdown request accepted')
        self.shutdown_requested = True
        if self.in_interruptible:
            logger.debug('Interruptible, raising ShutdownRequested')
            raise ShutdownRequested()

    @contextmanager
    def interruptible(self):
        self.maybe_shutdown()
        self.in_interruptible += 1
        # The signal handler may now raise ShutdownRequested
        try:
            yield
        finally:
            self.in_interruptible -= 1

    def maybe_shutdown(self):
        if self.shutdown_requested:
            logger.debug('Raising ShutdownRequested')
            raise ShutdownRequested()


class Maintenance:
    def __init__(self):
        self.last_run_at = None
        self.key = RedisKey('last_maintenance')

    def run_if_neccessary(self):
        if (self.last_run_at and
                (utcnow() - self.last_run_at) < timedelta(seconds=settings.MAINTENANCE_FREQ)):
            return
        # The cleanup tasks are not safe to run in paralell, so use this lock
        # to ensure that only one worker runs them.
        if connection.setnx(self.key, utcformat(utcnow())):
            connection.expire(self.key, settings.MAINTENANCE_FREQ)
            self.run()

        redis_value = connection.get(self.key)
        # might have expired between a failed SETNX and the GET
        if redis_value:
            # Use min to limit impact of workers with incorrect time
            self.last_run_at = min(utcnow(), utcparse(redis_value.decode()))

    def run(self):
        registry_maintenance()


class WorkHorse(multiprocessing.Process):
    def __init__(self, task, worker_connection):
        super().__init__(daemon=True)
        self.task = task
        self.worker_connection = worker_connection

    def run(self):
        ps = PostponeShutdown()
        ps.activate()
        self.setup_signal_handler()
        logger.debug("WorkHorse started")
        self.worker_connection.send(True)

        @contextmanager
        def shutdown_cm():
            ps.deactivate()
            try:
                yield
            finally:
                self.ignore_shutdown_signal()

        outcome = self.task.execute(shutdown_cm=shutdown_cm())
        self.worker_connection.send(outcome)
        logger.debug("WorkHorse finished")

    def setup_signal_handler(self):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGUSR1, self.request_stop)

    def ignore_shutdown_signal(self):
        signal.signal(signal.SIGUSR1, signal.SIG_IGN)

    def request_stop(self, signum, frame):
        logger.debug("WorkHorse received shutdown request")
        self.ignore_shutdown_signal()
        PostponeShutdown.trigger_shutdown()

    def send_signal(self, sig):
        os.kill(self.pid, sig)


class TWorker:
    def __init__(self, queues=['default']):
        id = str(uuid.uuid4())
        self.worker = Worker(id,
                             queues=[Queue(n) for n in queues],
                             description=f'TestWorker-{id}')
        self.failed = []
        self.succeeded = []

    def run(self, raise_on_failure=True):
        self.worker.startup()
        i = 0
        while True:
            for queue in self.worker.queues:
                task = queue.dequeue(self.worker)
                if task:
                    break
            else:
                break

            i += 1
            self.worker.start_task(task)
            outcome = task.execute()
            self.worker.end_task(task, outcome)

            if outcome.outcome == 'success':
                self.succeeded.append(task)
            elif outcome.outcome == 'failure':
                self.failed.append(task)
                if raise_on_failure:
                    raise RuntimeError(f"Task {task.description!r} failed\n" + task.error_message)
            else:
                raise RuntimeError("Unexpected task outcome")
        return i



import logging
from operator import attrgetter

from .conf import RedisKey, connection, settings
from .exceptions import WorkerDoesNotExist
from .queue import Queue
from .registries import worker_registry
from .task import Task
from .utils import atomic_pipeline, enum, utcformat, utcnow, utcparse

logger = logging.getLogger(__name__)

WorkerState = enum(
    'WorkerState',
    IDLE='idle',
    BUSY='busy',
    DEAD='dead',
)


class Worker:
    @classmethod
    def all(cls):
        return list(sorted((cls.fetch(id) for id in worker_registry.get_worker_ids()),
                           key=attrgetter('description')))

    @classmethod
    def fetch(cls, id):
        return cls(fetch_id=id)

    def __init__(self, id=None, *, description=None, queues=None,
                 fetch_id=None):
        self.id = id or fetch_id
        self.key = RedisKey('worker:' + self.id)
        self.task_key = RedisKey('worker_task:' + self.id)

        if fetch_id:
            self.refresh()
            return

        self.description = description or f"Worker {id}"
        self.state = None
        self.queues = queues
        if not queues:
            raise ValueError("Worker needs queues")
        self.current_task_id = None
        self.started_at = None
        self.shutdown_at = None

    def refresh(self):
        with connection.pipeline() as pipeline:
            pipeline.hgetall(self.key)
            pipeline.lrange(self.task_key, 0, -1)
            obj, task_id = pipeline.execute()

        if not obj:
            raise WorkerDoesNotExist(self.id)
        assert len(task_id) < 2
        self.current_task_id = task_id[0].decode() if task_id else None

        obj = {k.decode(): v.decode() for k, v in obj.items()}
        self.state = obj['state']
        self.description = obj['description']
        self.queues = [Queue(q) for q in obj['queues'].split(',')]
        for k in ['started_at', 'shutdown_at']:
            setattr(self, k, utcparse(obj[k]) if obj.get(k) else None)

    @atomic_pipeline
    def _save(self, fields=None, *, pipeline):
        string_fields = ['description', 'state']
        date_fields = ['started_at', 'shutdown_at']
        special_fields = ['queues', 'current_task_id']
        if fields is None:
            fields = string_fields + date_fields + special_fields

        deletes = []
        store = {}
        for field in fields:
            value = getattr(self, field)
            if field == 'queues':
                store['queues'] = ','.join(q.name for q in self.queues)
            elif field == 'current_task_id':
                pipeline.delete(self.task_key)
                if value:
                    pipeline.lpush(self.task_key, value)
            elif value is None:
                deletes.append(field)
            elif field in string_fields:
                store[field] = value
            elif field in date_fields:
                store[field] = utcformat(value)
            else:
                raise AttributeError(f'{field} is not a valid attribute')
        if deletes:
            pipeline.hdel(self.key, *deletes)
        if store:
            pipeline.hmset(self.key, store)

    def heartbeat(self):
        """Send a heartbeat.

        Raises WorkerDoesNotExist if the registry considers this worker as dead"""
        worker_registry.heartbeat(self)

    @atomic_pipeline
    def startup(self, *, pipeline):
        logger.info(f'Worker {self.description} [{self.id}] started')
        self.state = WorkerState.IDLE
        self.started_at = utcnow()
        worker_registry.add(self, pipeline=pipeline)
        self._save(pipeline=pipeline)

    @atomic_pipeline
    def start_task(self, task, *, pipeline):
        assert self.state == WorkerState.IDLE
        # The task should be assigned by the queue
        assert self.current_task_id == task.id
        self.state = WorkerState.BUSY
        self._save(['state'], pipeline=pipeline)
        task.set_running(self, pipeline=pipeline)

    @atomic_pipeline
    def end_task(self, task, outcome, *, pipeline):
        assert self.state == WorkerState.BUSY
        assert self.current_task_id == task.id
        self.state = WorkerState.IDLE
        self.current_task_id = None
        self._save(['state', 'current_task_id'], pipeline=pipeline)
        task.handle_outcome(outcome, pipeline=pipeline)

    @atomic_pipeline
    def shutdown(self, *, pipeline):
        logger.info(f'Worker {self.description} [{self.id}] shut down')
        assert self.state == WorkerState.IDLE
        worker_registry.remove(self, pipeline=pipeline)
        self.state = WorkerState.DEAD
        self.shutdown_at = utcnow()
        self._save(['state', 'shutdown_at'], pipeline=pipeline)
        pipeline.expire(self.key, settings.DEAD_WORKER_TTL)

    @atomic_pipeline
    def died(self, *, pipeline):
        logger.warning(f'Worker {self.description} [{self.id}] died')
        worker_registry.remove(self, pipeline=pipeline)
        self.shutdown_at = utcnow()
        if self.current_task_id:
            task = Task.fetch(self.current_task_id)
            task.handle_worker_death(pipeline=pipeline)
            self.current_task_id = None
        self.state = WorkerState.DEAD
        self._save(['state', 'current_task_id', 'shutdown_at'], pipeline=pipeline)
        pipeline.expire(self.key, settings.DEAD_WORKER_TTL)

    def fetch_current_task(self):
        """Returns the currently executing task."""
        if self.current_task_id:
            return Task.fetch(self.current_task_id)



import bisect
import datetime
from collections import namedtuple

import pytz
from pytz.tzinfo import DstTzInfo

Transition = namedtuple('Transition', [
    'start', 'end',
    'utc_start', 'utc_end',
    'old_utcoffset', 'new_utcoffset'])


class DstSmearingTz:
    def __init__(self, name):
        self._transition_times = []
        self._transition_times_utc = []
        self._transition_infos = []

        self.tz = pytz.timezone(name)
        if isinstance(self.tz, DstTzInfo):
            self._load_transitions()

    def _load_transitions(self):
        prev_utcoffset = None
        HOUR = datetime.timedelta(hours=1)
        for transition_time, (utcoffset, dstoffset, tzname) in \
                zip(self.tz._utc_transition_times, self.tz._transition_info):
            if prev_utcoffset is not None:
                if abs((prev_utcoffset - utcoffset).total_seconds()) == 3600:
                    if prev_utcoffset < utcoffset:
                        start = transition_time + prev_utcoffset - HOUR
                        end = transition_time + utcoffset + HOUR
                    else:
                        start = transition_time + utcoffset
                        end = transition_time + prev_utcoffset
                    transition = Transition(
                        start=start, end=end,
                        utc_start=transition_time - HOUR, utc_end=transition_time + HOUR,
                        old_utcoffset=prev_utcoffset, new_utcoffset=utcoffset,
                    )
                    self._transition_times.append(start)
                    self._transition_times.append(end)
                    self._transition_times_utc.append(transition.utc_start)
                    self._transition_times_utc.append(transition.utc_end)
                    self._transition_infos.append((True, transition))
                    self._transition_infos.append((False, transition))
            prev_utcoffset = utcoffset

    def from_utc(self, utc):
        if utc.utcoffset() is None or utc.utcoffset().total_seconds() != 0:
            raise ValueError('Datetime is not in utc')
        if not self._transition_infos:
            return utc.astimezone(self.tz).replace(tzinfo=None)
        utc = utc.replace(tzinfo=None)
        idx = bisect.bisect_right(self._transition_times_utc, utc) - 1
        in_trans, transition = self._transition_infos[idx]
        if not in_trans:
            return utc + transition.new_utcoffset
        else:
            percentage = (utc - transition.utc_start) / (transition.utc_end - transition.utc_start)
            utcoffset = (transition.old_utcoffset * (1 - percentage) +
                         transition.new_utcoffset * percentage)
            return utc + utcoffset

    def to_utc(self, dt):
        if dt.tzinfo is not None:
            raise ValueError('Not naive datetime (tzinfo is already set)')
        if not self._transition_infos:
            return self.tz.localize(dt).astimezone(pytz.UTC)
        idx = bisect.bisect_right(self._transition_times, dt) - 1
        in_trans, transition = self._transition_infos[idx]
        if not in_trans:
            utc = dt - transition.new_utcoffset
        else:
            percentage = (dt - transition.start) / (transition.end - transition.start)
            utcoffset = (transition.old_utcoffset * (1 - percentage) +
                         transition.new_utcoffset * percentage)
            utc = dt - utcoffset
        return utc.replace(tzinfo=pytz.UTC)



import importlib
import os
from types import SimpleNamespace

import redis

from redis_tasks import defaults

from .utils import LazyObject, import_attribute

ENVIRONMENT_VARIABLE = "RT_SETTINGS_MODULE"


class Settings:
    def __init__(self):
        self._initialized = False

    def _configure_from_env(self, setting_name=None):
        settings_module = os.environ.get(ENVIRONMENT_VARIABLE)
        if not settings_module:
            raise Exception(
                f"redis_tasks settings are not configured. "
                "You must either define the environment variable "
                f"{ENVIRONMENT_VARIABLE} or call settings.configure() before "
                "accessing redis_tasks.")

        mod = importlib.import_module(settings_module)
        self._setup(mod)

    def _setup(self, settings_module):
        for setting in dir(defaults):
            if setting.isupper():
                setattr(self, setting, getattr(defaults, setting))

        for setting in dir(settings_module):
            if setting.isupper():
                setattr(self, setting, getattr(settings_module, setting))

        self._initialized = True

    def __getattr__(self, name):
        if not self._initialized:
            self._configure_from_env(setting_name=name)
        return self.__dict__[name]

    def configure(self, settings):
        if self._initialized:
            raise RuntimeError('Settings already configured.')
        self._setup(settings)

    def configure_from_dict(self, dct):
        self.configure(SimpleNamespace(**dct))


settings = Settings()


class RTRedis(redis.StrictRedis):
    RESPONSE_CALLBACKS = redis.StrictRedis.RESPONSE_CALLBACKS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_response_callback('EXISTS', int)

    def pipeline(self, transaction=True, shard_hint=None):
        return RTPipeline(self.connection_pool, self.response_callbacks, transaction, shard_hint)

    def exists(self, *keys):
        return self.execute_command('EXISTS', *keys)

    def ftime(self):
        seconds, microseconds = self.time()
        return seconds + microseconds * 10**-6

    def zadd(self, name, items, nx=False, xx=False, ch=False, incr=False):
        if nx and xx:
            raise redis.RedisError("ZADD can't use both NX and XX modes")
        pieces = []
        if nx:
            pieces.append('NX')
        if xx:
            pieces.append('XX')
        if ch:
            pieces.append('CH')
        if incr:
            pieces.append('INCR')
        for k, v in items.items():
            pieces.extend([v, k])
        return self.execute_command('ZADD', name, *pieces)


class RTPipeline(redis.client.BasePipeline, RTRedis):
    pass


@LazyObject
def connection():
    return RTRedis.from_url(settings.REDIS_URL)


@LazyObject
def task_middleware():  # TODO: test
    def middleware_constructor(class_path):
        return import_attribute(class_path)

    return [middleware_constructor(x) for x in settings.MIDDLEWARE]


class RedisKey:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return settings.REDIS_PREFIX + ':' + self.name

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.name == other.name



import uuid

from redis_tasks import Queue
from redis_tasks.conf import RedisKey, connection
from redis_tasks.utils import atomic_pipeline, serialize, deserialize


def chain(members):
    graph = TaskGraph()
    tail = None
    for member in members:
        node = graph.add_task(member)
        if tail:
            graph.add_dependency(tail, node)
        tail = node
    return graph


class TaskGraph:
    def __init__(self, id=None):
        self.id = id or str(uuid.uuid4())
        self.key = RedisKey(f"ston.graph:{self.id}")
        self.nodes = []
        self.edges = set()

    def add_task(self, task):
        if not isinstance(task["func"], str):
            task["func"] = '{0}.{1}'.format(task["func"].__module__, task["func"].__name__)
        node = GraphNode(task)
        self.nodes.append(node)
        return node

    def add_dependency(self, before, after):
        self.edges.add((before, after))

    @atomic_pipeline
    def save(self, *, pipeline):
        if self.nodes:
            pipeline.set(self.key, serialize({
                'nodes': {id(node): (node.task, node.task_id) for node in self.nodes},
                'edges': [(id(a), id(b)) for (a, b) in self.edges]
            }), ex=7 * 24 * 60 * 60)
        else:
            pipeline.delete(self.key)

    def reload(self):
        dct = deserialize(connection.get(self.key))
        nodes = {id: GraphNode(*data) for id, data in dct["nodes"].items()}
        self.nodes = list(nodes.values())
        self.edges = {(nodes[a], nodes[b]) for (a, b) in dct["edges"]}

    def lock(self):
        return connection.lock(f"{self.key}:lock", timeout=10)

    def mark_done(self, task_id):
        self.nodes = [n for n in self.nodes if n.task_id != task_id]
        self.edges = {(a, b) for (a, b) in self.edges if a.task_id != task_id}

    @atomic_pipeline
    def enqueue_ready(self, *, pipeline):
        blocked = {b for (a, b) in self.edges}
        # We want to keep the order of the nodes
        ready = [node for node in self.nodes
                 if not node.task_id and node not in blocked]
        for node in ready:
            queue = Queue(node.task.pop('queue', "default"))
            task = queue.enqueue_call(**node.task, pipeline=pipeline)
            node.task = None
            task.meta["ston.graph"] = self.id
            task.save_meta(pipeline=pipeline)
            node.task_id = task.id

    @atomic_pipeline
    def enqueue(self, *, pipeline):
        if not GraphMiddleware.is_installed():
            raise Exception(
                f"Please add {GraphMiddleware.__module__}.{GraphMiddleware.__name__} "
                "to your MIDDLEWARE")
        self.enqueue_ready(pipeline=pipeline)
        self.save(pipeline=pipeline)


class GraphNode:
    def __init__(self, task=None, task_id=None):
        self.task = task
        self.task_id = task_id


class GraphMiddleware:
    def process_outcome(self, task, *exc_info):
        graph_id = task.meta.get("ston.graph")
        if not graph_id:
            return
        graph = TaskGraph(graph_id)
        with graph.lock():
            # TODO: handle missing graph
            graph.reload()
            graph.mark_done(task.id)
            with connection.pipeline() as pipe:
                graph.enqueue_ready(pipeline=pipe)
                graph.save(pipeline=pipe)
                pipe.execute()

    @classmethod
    def is_installed(cls):
        from redis_tasks.conf import task_middleware
        print(task_middleware)
        return any(issubclass(x, cls) for x in task_middleware)



import logging
import sys
import threading
import traceback
import uuid
from contextlib import ExitStack

import redis_tasks

from .conf import RedisKey, connection, settings, task_middleware
from .exceptions import (
    InvalidOperation, TaskAborted, TaskDoesNotExist, WorkerShutdown)
from .registries import failed_task_registry, finished_task_registry
from .utils import (
    LazyObject, atomic_pipeline, deserialize, enum, generate_callstring,
    import_attribute, serialize, utcformat, utcnow, utcparse)

logger = logging.getLogger(__name__)

TaskStatus = enum(
    'TaskStatus',
    QUEUED='queued',
    FINISHED='finished',
    FAILED='failed',
    CANCELED='canceled',
    RUNNING='running',
)


def redis_task(*args, **kwargs):
    def decorator(f):
        # Constructing TaskProperties instances initializes the settings.
        # We do not want to do this on import of user modules, so be lazy here.
        f._redis_task_properties = LazyObject(lambda: TaskProperties(*args, **kwargs))
        return f

    return decorator


class TaskStack():
    def __init__(self):
        self.local = threading.local()

    def push(self, task):
        stack = getattr(self.local, "stack", [])
        stack.append(task)
        self.local.stack = stack

    def pop(self):
        return self.local.stack.pop()

    def peek(self):
        stack = getattr(self.local, "stack", [])
        if not stack:
            return None
        return stack[-1]


task_stack = TaskStack()


def get_current_task():
    return task_stack.peek()


class TaskProperties:
    def __init__(self, reentrant=False, timeout=None):
        self.reentrant = reentrant
        self.timeout = timeout or settings.DEFAULT_TASK_TIMEOUT


class TaskOutcome:
    def __init__(self, outcome, message=None):
        assert outcome in ['success', 'failure', 'requeue']
        self.outcome = outcome
        self.message = message

    def __repr__(self):
        args = [self.outcome]
        if self.message:
            args.append(f'message={self.message!r}')
        return 'TaskOutcome({})'.format(", ".join(args))


class Task:
    def __init__(self, func=None, args=None, kwargs=None, *, fetch_id=None):
        if fetch_id:
            self.id = fetch_id
            self.refresh()
            return

        self.id = str(uuid.uuid4())

        try:
            if isinstance(func, str):
                self.func_name = func
                func = self._get_func()
            else:
                self.func_name = '{0}.{1}'.format(func.__module__, func.__name__)
                assert self._get_func() == func
        except Exception as e:
            raise ValueError(f'The given task function {self.func_name!r} is not importable') from e
        if not callable(func):
            raise ValueError(f'The given task function {self.func_name!r} is not callable')

        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        if not isinstance(args, (tuple, list)):
            raise TypeError(f'{args!r} is not a valid args list')
        if not isinstance(kwargs, dict):
            raise TypeError(f'{kwargs!r} is not a valid kwargs dict')
        self.args = args
        self.kwargs = kwargs

        self.error_message = None

        self.description = generate_callstring(self.func_name, self.args, self.kwargs)
        self.status = None
        self.origin = None
        self.meta = {}

        self.enqueued_at = None
        self.started_at = None
        self.ended_at = None
        self.aborted_runs = []

    @classmethod
    def fetch(cls, id):
        return cls(fetch_id=id)

    @property
    def key(self):
        return self.key_for(self.id)

    @classmethod
    def key_for(cls, task_id):
        return RedisKey('task:' + task_id)

    @atomic_pipeline
    def enqueue(self, queue, *, pipeline):
        assert self.status is None
        logger.info(f"Task {self.description} [{self.id}] enqueued")
        self.status = TaskStatus.QUEUED
        self.origin = queue.name
        self.enqueued_at = utcnow()
        self._save(pipeline=pipeline)
        queue.push(self, pipeline=pipeline)

    @atomic_pipeline
    def requeue(self, *, pipeline):
        assert self.status == TaskStatus.RUNNING
        logger.info(f"Task {self.description} [{self.id}] requeued")
        redis_tasks.Queue(self.origin).push(self, at_front=True, pipeline=pipeline)
        self.status = TaskStatus.QUEUED
        self.aborted_runs.append((self.started_at, utcnow()))
        self.started_at = None
        self._save(['status', 'aborted_runs', 'started_at'], pipeline=pipeline)

    @atomic_pipeline
    def set_running(self, worker, *, pipeline):
        assert self.status == TaskStatus.QUEUED
        logger.info(f"Task {self.description} [{self.id}] started")
        self.status = TaskStatus.RUNNING
        self.started_at = utcnow()
        self._save(['status', 'started_at'], pipeline=pipeline)

    @atomic_pipeline
    def set_finished(self, *, pipeline):
        assert self.status == TaskStatus.RUNNING
        logger.info(f"Task {self.description} [{self.id}] finished")
        finished_task_registry.add(self, pipeline=pipeline)
        self.status = TaskStatus.FINISHED
        self.ended_at = utcnow()
        self._save(['status', 'ended_at'], pipeline=pipeline)

    @atomic_pipeline
    def set_failed(self, error_message, *, pipeline):
        assert self.status == TaskStatus.RUNNING
        logger.info(f"Task {self.description} [{self.id}] failed")
        failed_task_registry.add(self, pipeline=pipeline)
        self.status = TaskStatus.FAILED
        self.error_message = error_message
        self.ended_at = utcnow()
        self._save(['status', 'error_message', 'ended_at'], pipeline=pipeline)

    @atomic_pipeline
    def handle_outcome(self, outcome, *, pipeline):
        if outcome.outcome == 'success':
            self.set_finished(pipeline=pipeline)
        elif outcome.outcome == 'failure':
            self.set_failed(outcome.message, pipeline=pipeline)
        elif outcome.outcome == 'requeue':
            self.requeue(pipeline=pipeline)

    @atomic_pipeline
    def handle_worker_death(self, *, pipeline):
        if self.status == TaskStatus.QUEUED:
            logger.debug(f"Task {self.description} [{self.id}] had its worker die. Reenqueuing.")
            # The worker died while moving the task
            redis_tasks.Queue(self.origin).push(self, at_front=True, pipeline=pipeline)
        elif self.status == TaskStatus.RUNNING:
            outcome = self.get_abort_outcome("Worker died")
            self.handle_outcome(outcome, pipeline=pipeline)
        else:
            raise Exception(f"Unexpected task status: {self.status}")

    def get_abort_outcome(self, message, *, may_requeue=True):
        if may_requeue and self.is_reentrant:
            return TaskOutcome('requeue')
        else:
            try:
                raise TaskAborted(message)
            except TaskAborted:
                exc_info = sys.exc_info()
            return self._generate_outcome(*exc_info, may_requeue=may_requeue)

    def cancel(self):
        queue = redis_tasks.Queue(name=self.origin)
        try:
            queue.remove_and_delete(self)
        except TaskDoesNotExist:
            raise InvalidOperation("Only enqueued jobs can be canceled")
        self.status = TaskStatus.CANCELED

    def _get_func(self):
        return import_attribute(self.func_name)

    def _get_properties(self):
        return getattr(self._get_func(), '_redis_task_properties', TaskProperties())

    @property
    def is_reentrant(self):
        try:
            return self._get_properties().reentrant
        except Exception:
            return TaskProperties().reentrant

    @property
    def timeout(self):
        try:
            return self._get_properties().timeout
        except Exception:
            return TaskProperties().timeout

    @property
    def queue(self):
        from .queue import Queue
        return Queue(self.origin)

    @classmethod
    @atomic_pipeline
    def delete_many(cls, task_ids, *, pipeline):
        if task_ids:
            pipeline.delete(*(cls.key_for(task_id) for task_id in task_ids))

    def refresh(self):
        key = self.key
        obj = {k.decode(): v for k, v in connection.hgetall(key).items()}
        if len(obj) == 0:
            raise TaskDoesNotExist('No such task: {0}'.format(key))

        self.func_name = obj['func_name'].decode()
        self.args = deserialize(obj['args'])
        self.kwargs = deserialize(obj['kwargs'])

        for key in ['status', 'origin', 'description', 'error_message']:
            setattr(self, key, obj[key].decode() if key in obj else None)

        for key in ['enqueued_at', 'started_at', 'ended_at']:
            setattr(self, key, utcparse(obj[key].decode()) if key in obj else None)

        self.meta = deserialize(obj['meta']) if obj.get('meta') else {}
        self.aborted_runs = deserialize(obj['aborted_runs']) if obj.get('aborted_runs') else []

    @atomic_pipeline
    def _save(self, fields=None, *, pipeline=None):
        string_fields = ['func_name', 'status', 'description', 'origin', 'error_message']
        date_fields = ['enqueued_at', 'started_at', 'ended_at']
        data_fields = ['args', 'kwargs', 'meta', 'aborted_runs']
        if fields is None:
            fields = string_fields + date_fields + data_fields

        deletes = []
        store = {}
        for field in fields:
            value = getattr(self, field)
            if value is None:
                deletes.append(field)
            elif field in string_fields:
                store[field] = value
            elif field in date_fields:
                store[field] = utcformat(value)
            elif field in data_fields:
                store[field] = serialize(value)
            else:
                raise AttributeError(f'{field} is not a valid attribute')
        if deletes:
            pipeline.hdel(self.key, *deletes)
        if store:
            pipeline.hmset(self.key, store)

    @atomic_pipeline
    def save_meta(self, *, pipeline=None):
        self._save(['meta'], pipeline=pipeline)

    def execute(self, *, shutdown_cm=ExitStack()):
        """Run the task using middleware.

        The `shutdown_cm` parameter is a context manager that will wrap the part
        of the execution in which `WorkerShutdown` is allowed to be raised.

        Returns a TaskOutcome."""
        task_stack.push(self)
        exc_info = (None, None, None)
        try:
            def run_task(*args, **kwargs):
                with shutdown_cm:
                    try:
                        func = self._get_func()
                    except Exception as e:
                        raise RuntimeError(
                            f"Failed to import task function {self.func_name}") from e
                    func(*args, **kwargs)

            def mw_wrapper(mwc, task, run):
                def mw_run(*args, **kwargs):
                    middleware = mwc()
                    if hasattr(middleware, "run_task"):
                        middleware.run_task(task, run, args, kwargs)
                    else:
                        run(*args, **kwargs)
                return mw_run

            for middleware_constructor in reversed(task_middleware):
                run_task = mw_wrapper(middleware_constructor, self, run_task)

            try:
                run_task(*self.args, **self.kwargs)
            except WorkerShutdown as e:
                raise TaskAborted("Worker shutdown") from e
        except Exception:
            exc_info = sys.exc_info()
        finally:
            task_stack.pop()

        return self._generate_outcome(*exc_info)

    def _generate_outcome(self, *exc_info, may_requeue=True):
        if may_requeue and isinstance(exc_info[1], TaskAborted) and self.is_reentrant:
            return TaskOutcome('requeue')
        for mwc in reversed(task_middleware):
            try:
                middleware = mwc()
                if not hasattr(middleware, "process_outcome"):
                    continue
                if middleware.process_outcome(self, *exc_info):
                    exc_info = (None, None, None)
            except Exception:
                exc_info = sys.exc_info()

        if not exc_info[0]:
            return TaskOutcome('success')
        else:
            exc_string = ''.join(traceback.format_exception(*exc_info))
            return TaskOutcome('failure', message=exc_string)

    def __repr__(self):
        return '<{0} {1}: {2}>'.format(
            self.__class__.__name__, self.id, self.description)



import datetime
import logging
import random
import signal
import threading
import time
import uuid
from contextlib import suppress

import croniter

from .conf import RedisKey, connection, settings
from .exceptions import TaskDoesNotExist
from .queue import Queue
from .smear_dst import DstSmearingTz
from .task import Task, TaskStatus
from .utils import (
    LazyObject, atomic_pipeline, decode_dict, utcformat, utcnow, utcparse)

logger = logging.getLogger(__name__)

local_tz = LazyObject(lambda: DstSmearingTz(settings.SCHEDULER_TIMEZONE))


class CrontabSchedule:
    def __init__(self, crontab):
        self.crontab = crontab

    def get_next(self, after):
        after = local_tz.from_utc(after)
        iter = croniter.croniter(self.crontab, after, ret_type=datetime.datetime)
        return local_tz.to_utc(iter.get_next())


crontab = CrontabSchedule


def once_per_day(time_str):
    hour, minute = time_str.split(':')
    return CrontabSchedule(f'{minute} {hour} * * *')


class PeriodicSchedule:
    def __init__(self, *, hours=0, minutes=0, seconds=0, start_at=None):
        self.interval = hours * 60 * 60 + minutes * 60 + seconds
        if start_at is None:
            start_at = random.uniform(0, self.interval)
        elif isinstance(start_at, str):
            hours, minutes = start_at.split(':')
            start_at = int(hours) * 60 * 60 + int(minutes) * 60
        self.start_at = start_at
        assert self.interval < 12 * 60 * 60
        assert self.start_at < 24 * 60 * 60

    def get_next(self, after):
        after = after.astimezone(local_tz.tz)
        midnight = local_tz.tz.localize(
            after.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None),
            is_dst=True)
        after_time = (after - midnight).total_seconds()
        if self.start_at > after_time:
            next_time = self.start_at
        else:
            since_start = after_time - self.start_at
            remaining = self.interval - since_start % self.interval
            next_time = after_time + remaining

        next = local_tz.tz.normalize(midnight + datetime.timedelta(seconds=next_time))
        if next.day != after.day:
            midnight = local_tz.tz.localize(
                after.replace(hour=23, minute=59, second=59, microsecond=0, tzinfo=None),
                is_dst=False) + datetime.timedelta(seconds=1)
            next = midnight + datetime.timedelta(seconds=self.start_at)
        return next.astimezone(datetime.timezone.utc)


run_every = PeriodicSchedule


class SchedulerEntry:
    def __init__(self, id, config):
        self.id = id
        self.key = RedisKey(f"schedule_entry:{self.id}")
        self.singleton = config.get('singleton', True)
        self.task_template = [config['task'],
                              config.get('args', ()), config.get('kwargs', {})]
        self.queue = Queue(config.get('queue', settings.SCHEDULER_QUEUE))
        self.schedule = config['schedule']

        self.last_save = None
        stored = decode_dict(connection.hgetall(self.key))
        prev_run = stored.get("prev_run")
        self.prev_run = utcparse(prev_run) if prev_run else utcnow()
        self.prev_task_id = stored.get("prev_task_id")

        self.next_run = self.schedule.get_next(self.prev_run)

        # Make sure task config is valid
        Task(*self.task_template)

    @atomic_pipeline
    def save(self, *, pipeline):
        pipeline.hset(self.key, "prev_run", utcformat(self.prev_run))
        if self.prev_task_id:
            pipeline.hset(self.key, "prev_task_id", self.prev_task_id)
        else:
            pipeline.hdel(self.key, "prev_task_id")
        ttl = max(24 * 60 * 60, settings.SCHEDULER_MAX_CATCHUP * 5)
        pipeline.expire(self.key, ttl)
        self.last_save = utcnow()

    @atomic_pipeline
    def process(self, now, *, pipeline):
        max_catchup = now - datetime.timedelta(seconds=settings.SCHEDULER_MAX_CATCHUP)
        self.next_run = self.schedule.get_next(max(max_catchup, self.prev_run))
        if self.next_run > now:
            if (not self.last_save or
                    (now - self.last_save).total_seconds() >= settings.SCHEDULER_MAX_CATCHUP):
                self.save(pipeline=pipeline)
            return

        self.prev_run = now

        if self.singleton:
            self.next_run = self.schedule.get_next(now)
            if self.is_enqueued():
                logger.info(f'Schedule entry "{self.id}" already enqueued or running, skipping')
            else:
                self.enqueue(pipeline=pipeline)
        else:
            while self.next_run <= now:
                self.enqueue(pipeline=pipeline)
                self.next_run = self.schedule.get_next(self.next_run)
        self.save(pipeline=pipeline)

    def is_enqueued(self):
        if self.prev_task_id:
            with suppress(TaskDoesNotExist):
                prev_task = Task.fetch(self.prev_task_id)
                if prev_task.status not in [TaskStatus.FINISHED, TaskStatus.FAILED]:
                    return True
        return False

    @atomic_pipeline
    def enqueue(self, *, pipeline):
        task = self.queue.enqueue_call(*self.task_template, pipeline=pipeline)
        self.prev_task_id = task.id
        return task


class Scheduler:
    def __init__(self):
        self.schedule = [SchedulerEntry(k, v) for k, v in settings.SCHEDULE.items()]
        self.shutdown_requested = threading.Event()

    def setup_signal_handler(self):
        def stop(signum, frame):
            logger.info('Initiating redis_tasks scheduler shutdown')
            self.shutdown_requested.set()

        signal.signal(signal.SIGINT, stop)
        signal.signal(signal.SIGTERM, stop)

    def run(self):
        if not self.schedule:
            logger.error("No schedule configured, nothing to do")
            return

        self.setup_signal_handler()

        HEARTBEAT_FREQ = 10
        with Mutex(timeout=HEARTBEAT_FREQ + 2) as mutex:
            logger.info('redis_tasks scheduler started')
            while not self.shutdown_requested.is_set():
                mutex.extend()
                now = utcnow()
                for entry in self.schedule:
                    entry.process(now)

                next_run = min(x.next_run for x in self.schedule)
                next_heartbeat = now + datetime.timedelta(seconds=HEARTBEAT_FREQ)
                wait_for = (utcnow() - min(next_run, next_heartbeat)).total_seconds()
                self.shutdown_requested.wait(wait_for)
        logger.info('redis_tasks scheduler shut down')


def scheduler_main():
    Scheduler().run()


class Mutex(object):
    expire_script = None

    def __init__(self, *, timeout):
        self.key = RedisKey('scheduler')
        self.timeout = timeout
        self.token = None

        # KEYS[1]: lock key, ARGS[1]: token, ARGS[2]: milliseconds
        # return 1 if the lock was held and the expire executed, otherwise 0
        self.expire_script = connection.register_script("""
            if redis.call('get', KEYS[1]) == ARGV[1] then
                return redis.call('pexpire', KEYS[1], ARGV[2])
            else
                return 0
            end""")

    def __enter__(self):
        if not self.acquire(wait=False):
            wait_for = self.timeout + 1
            logger.warning("Found signs of an already running scheduler instance, "
                           f"waiting {wait_for} seconds for it to disappear")
            try:
                self.acquire(wait=wait_for)
            except TimeoutError:
                raise RuntimeError("redis_tasks scheduler already running")
        return self

    def __exit__(self, *exc):
        if self.token:
            self.expire_script(keys=[self.key], args=[self.token, 0])
            self.token = None

    def acquire(self, wait=None):
        token = str(uuid.uuid1()).encode()
        stop_trying_at = time.time() + wait
        while True:
            acquired = connection.set(self.key, token, nx=True, px=int(self.timeout * 1000))
            if acquired:
                self.token = token
                return True
            elif not wait:
                return False
            elif time.time() > stop_trying_at:
                raise TimeoutError
            time.sleep(0.1)

    def extend(self):
        if not self.expire_script(keys=[self.key], args=[self.token, int(self.timeout * 1000)]):
            raise RuntimeError("Cannot refresh a lock that's no longer owned")



import datetime
import importlib
import operator
import pickle
from functools import wraps

from .exceptions import DeserializationError


def import_attribute(name):
    module_name, attribute = name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, attribute)


def generate_callstring(func_name, args, kwargs):  # TODO: test
    arg_list = [repr(arg) for arg in args]
    arg_list += [f'{k}={v!r}' for k, v in kwargs.items()]
    args = ', '.join(arg_list)
    return f'{func_name}({args})'


def utcnow():
    return datetime.datetime.now(datetime.timezone.utc)


def utcformat(dt):
    assert dt.tzinfo and dt.tzinfo.utcoffset(dt).total_seconds() == 0
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


def utcparse(string):
    parsed = datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%SZ')
    return parsed.replace(tzinfo=datetime.timezone.utc)


def enum(name, *sequential, **named):
    values = dict(zip(sequential, range(len(sequential))), **named)
    return type(str(name), (), values)


def one(iterable):
    it = iter(iterable)
    try:
        value = next(it)
    except StopIteration:
        raise ValueError('too few items in iterable (expected 1)')
    try:
        next(it)
    except StopIteration:
        return value
    else:
        raise ValueError('too many items in iterable (expected 1)')


def decode_list(lst):
    return [x.decode() for x in lst]


def decode_dict(dct):
    return {k.decode(): v.decode() for k, v in dct.items()}


def is_serializable(obj):
    if obj is None:
        return True
    if isinstance(obj, (int, str, float, bool, datetime.datetime)):
        return True
    if isinstance(obj, (tuple, list, set)):
        return all(is_serializable(x) for x in obj)
    if isinstance(obj, dict):
        return all(is_serializable(k) and is_serializable(v) for k, v in obj.items())
    return False


def serialize(obj):
    if not is_serializable(obj):
        raise ValueError("Passed object contains non-serializable values")
    return pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)


def deserialize(bytes_obj):
    try:
        obj = pickle.loads(bytes_obj)
    except Exception as e:
        raise DeserializationError('Could not unpickle', bytes_obj) from e
    return obj


def atomic_pipeline(f):
    from .conf import connection

    @wraps(f)
    def pipeline_wrapper(*args, pipeline=None, **kwargs):
        if pipeline is None:
            with connection.pipeline() as pipe:
                ret = f(*args, pipeline=pipe, **kwargs)
                pipe.execute()
                return ret
        else:
            return f(*args, pipeline=pipeline, **kwargs)

    return pipeline_wrapper


empty = object()


def new_method_proxy(func):
    def inner(self, *args, **kwargs):
        __tracebackhide__ = True
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args, **kwargs)
    return inner


class LazyObject:
    def __init__(self, func):
        self.__dict__['_wrapped'] = empty
        self.__dict__['_setupfunc'] = func

    def _setup(self):
        self.__dict__['_wrapped'] = self._setupfunc()
        del self.__dict__['_setupfunc']

    __getattr__ = new_method_proxy(getattr)
    __setattr__ = new_method_proxy(setattr)
    __delattr__ = new_method_proxy(delattr)

    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)

    __dir__ = new_method_proxy(dir)
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __reversed__ = new_method_proxy(reversed)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)

    def __repr__(self):
        if self._wrapped is empty:
            repr_attr = self._setupfunc
        else:
            repr_attr = self._wrapped
        return f'<{type(self).__name__}: {repr_attr!r}>'
